# Implement methods use for handle with table records like
# 1. Methods in as_json
# 2.
module Pureapi::Model
  # Define constants use for pure api
  module Pagination
    DEFAULT_PER_PAGE = 10
    INFINITE_PER_PAGE = 'infinite'.freeze
  end

  COMPARISON_OPERATORS = {
    eq: '=',
    ne: '<>',
    gt: '>',
    lt: '<',
    gte: '>=',
    lte: '<=',
    like: 'LIKE',
    nlike: 'NOT LIKE',
    regex: 'REGEXP',
  }

  NULL_OPERATORS = {
    'null' => nil,
    'blank' => [nil, ''],
  }

  LOGICAL_OPERATORS = {
    in: 'IN'
  }

  COMPARISON_OPERATORS_INVERT = COMPARISON_OPERATORS.invert
  LOGICAL_OPERATORS_INVERT = LOGICAL_OPERATORS.invert
  NULL_OPERATOR_KEYS = NULL_OPERATORS.keys

  # New version of +COMPARISON_OPERATORS+ & others
  module Operators
    EQ = { id: :eq, is: '=' }
    NE = { id: :ne, is: '<>' }
    GT = { id: :gt, is: '>' }
    LT = { id: :lt, is: '<' }
    GTE = { id: :gte, is: '>=' }
    LTE = { id: :lte, is: '<=' }
    LIKE = { id: :like, is: 'LIKE' }
    NLIKE = { id: :nlike, is: 'NOT LIKE' }
    REGEXP = { id: :regexp, is: 'REGEXP' }
    IN = { id: :in, is: 'IN', permit: [] }
    ISNIL = { id: :isnil, is: 'IS NULL' }
    ISNNIL = { id: :isnnil, is: 'IS NOT NULL' }
  end

  REFLECT_OPERATORS = Operators.constants.inject({}) do |memo, const_name|
    item = Operators.const_get(const_name)
    memo[item[:id]] = item[:is]
    memo
  end

  PERMIT_OPERATORS = Operators.constants.inject({}) do |memo, const_name|
    item = Operators.const_get(const_name)
    memo[item[:id]] = item[:permit]
    memo
  end

  # return default array contains methods for as_json
  # Eg. [:status_name, :path_detail, :restrictions, :channel_course_code]
  def json_methods
    []
  end

  # return default criterias for advanced_search
  # Eg.
  # User .where({}) to fix case return self
  # def advanced_search(advanced_params)
  #   criterias = self

  #   # search by list course_ids
  #   criterias = advanced_params[:course_ids].blank? ?
  #     criterias : criterias.where(course_id: advanced_params[:course_ids])

  #   # search by list channel_ids
  #   criterias = advanced_params[:channel_ids].blank? ?
  #     criterias : criterias.where(channel_id: advanced_params[:channel_ids])

  #   # search by code
  #   criterias = advanced_params[:code].blank? ?
  #     criterias : criterias.where(
  #       "`channel_courses`.`code` LIKE ?",
  #       "%#{advanced_params[:code]}%"
  #     )

  #   return criterias.where({})
  # end
  def advanced_search(advanced_params)
    return self.where({})
  end

  # Provide search function 2 type: normal and advanced
  # Required setup advanced_search methods
  #
  # ActionController::Parameters +_params+, contain fields use .where()
  # ActionController::Parameters +_advanced_params+, detail in advanced_search method
  # Or
  # Hash +_params+, contain fields use .where()
  # Hash +_advanced_params+, detail in advanced_search method
  def full_search(params, advanced_params)
    _params = params.delete_if { |k, v| v.blank? }

    criterias = self
    criterias = _params.blank? ? criterias : criterias.where(_params.to_h)
    criterias = advanced_params.blank? ? criterias : criterias.advanced_search(advanced_params.to_h)

    return criterias.where({})
  end

  # Comparsion operator conditions method
  # Array +params+ contains hash = {f, o, v}
  # :f is column name
  # :o is comparsion operator
  # :v is value
  def compconds(params = [])
    criterias = self

    params.each do |param|
      if param[:o] == COMPARISON_OPERATORS[:eq]
        criterias = criterias.where(param[:f] => param[:v])
      else
        criterias = criterias.where("`#{self.table_name}`.`#{param[:f]}` #{param[:o]} ?", param[:v])
      end
    end

    return criterias.where({})
  end

  # Comparsion operator conditions method
  # Array +params+ contains hash = {f, o, v}
  # :f is column name
  # :o is comparsion operator
  # :v is value
  def logicconds(params = [])
    criterias = self

    params.each do |param|
      if param[:o] == LOGICAL_OPERATORS[:in]
        criterias = criterias.where(param[:f] => param[:v])
      end
    end

    return criterias.where({})
  end

  # Define array of columns use for comparison conditions
  # E.g [:id, :name, :channel_course_id, :status]
  def compcond_columns
    [:id]
  end

  # Define array of columns use for in conditions (logical)
  # E.g [:id, :name, :channel_course_id, :status]
  def incond_columns
    [:id]
  end

  # Define default includes of active records
  # E.g [:channel, :course]
  def default_includes
    []
  end

  # Define default fields use for asjson on includes
  # E.g
  # {
  #   channel: {only: [:id, :code, :name], include: Channel.asjson_fields},
  #   course: {only: [:id, :code, :name]}
  # }
  def asjson_fields
    {}
  end

  def json_method_includes
    {}
  end

  # Define default fields use for as_jsons
  def default_onlyasjsons
    self.column_names.map(&:to_sym)
  end

  # Define relation entities can be included
  # Example:
  #   >> {User => [:manager], Department => [:department]}
  def include_entities
    {}
  end

  # Require params is hash contain Integer :page, :per_page
  def paginate(params)
    self.limit(params[:per_page]).offset((params[:page] - 1) * params[:per_page])
  end

  # Get all includes of relations in model
  def get_includes nested_entities
    _results = {}
    _entities = include_entities

    (_entities.keys - nested_entities).each do |_c|
      _entities[_c].each do |_r|
        _results[_r] = _c._json_fields(nested_entities)
      end
    end

    _results
  end

  # Return all available values around model
  def _json_fields nested_entities = []
    nested_entities << self

    {
      only: default_onlyasjsons,
      methods: json_methods,
      include: get_includes(nested_entities),
    }
  end

  # Static of +_json_fields+
  def as_json_fields
    @as_json_fields ||= self._json_fields
  end

  ### Begin methods for feature compconds filter
  # Return list of all belongs_to => class_name
  # Example:
  #   >> {ad: Ad, lead_level: LeadLevel, course: Course}
  def of_entities
    {}
  end

  # Full compcond regexp for key
  # Example:
  #   >> id.gte
  def compcond_regexp
    @compcond_regexp ||= /^(#{compcond_columns.join("|")})\.(#{REFLECT_OPERATORS.keys.join("|")})$/
  end

  # Compcond regexp for key
  # Example:
  #   >> id
  def compcond_regexp2nd
    @compcond_regexp2nd ||= /^(#{compcond_columns.join("|")})$/
  end

  # Return matchs of key with +compcond_regexp+ || +compcond_regexp2nd+
  def compcond_match(str)
    _matchs = compcond_regexp.match(str) || compcond_regexp2nd.match(str)

    if _matchs
      {
        f: _matchs[1],
        o: _matchs[2] ? _matchs[2].to_sym : :eq,
      }
    else
      nil
    end
  end

  # Return all fields can use for search of model
  def _search_fields
    _results = {only: compcond_columns, include: {}}

    of_entities.each do |entity, _c|
      _results[:include][entity] = _c._search_fields
    end

    _results
  end

  # Static of +_search_fields+
  def compcond_fields
    @compcond_fields ||= self._search_fields
  end

  # Search by params of attributes of model
  def _self_search only_params
    criterias = self

    only_params.each do |filter|
      case filter[:o]
      when Operators::EQ[:id], Operators::IN[:id]
        criterias = criterias.where(filter[:f] => filter[:v])
      when Operators::ISNIL[:id], Operators::ISNNIL[:id]
        criterias = criterias.where("`#{self.table_name}`.`#{filter[:f]}` #{REFLECT_OPERATORS[filter[:o]]}")
      else
        criterias = criterias.where("`#{self.table_name}`.`#{filter[:f]}` #{REFLECT_OPERATORS[filter[:o]]} ?", filter[:v])
      end
    end

    return criterias.where({})
  end

  # Search by params of relations
  def _include_search include_params
    criterias = self

    include_params.each do |entity, query|
      criterias = criterias.where(entity => of_entities[entity].compcond_search(query)) if of_entities[entity]
    end

    return criterias.where({})
  end

  # Collaborate +_self_search+ & +_include_search+
  def compcond_search params
    criterias = self

    criterias = criterias._self_search(params[:only]) if params[:only].is_a?(Array)
    criterias = criterias._include_search(params[:include]) if params[:include].is_a?(Hash)

    return criterias.where({})
  end
  ### End methods for feature compconds filter
end
